--!NEEDS:splayables.lua

SPlayer1 = Core.class(SPlayables) -- don't need to put in lists

function SPlayer1:init(xcworld, xparams, xparams2)
	self.cworld = xcworld
	-- system
	self.isplayer1 = true
	self.isactive = true
	-- self
	self.scalex = 1
	self.scaley = self.scalex
	self.flip = 1
	self.defaultmass = 0.45
	self.currmass = self.defaultmass
	--
	self.totallives = 3 -- 3
	self.currlives = self.totallives
	self.totalhealth = 5 -- 5
	self.currhealth = self.totalhealth
	--
	self.speed = 8.7 -- 8.8, 9
	self.jumpspeed = 46
	self.vx = 0
	self.vy = 0
	self.inputbuffer = 5 -- 4, new 20230731
	self.currinputbuffer = 0 -- new 20230731
	self.coyotetimer = 12
	self.currcoyotetimer = 0
	-- animations
	self.hasanims = true -- allows to have animations
	self.hasLRanims = true -- allows to have L and R animations (note: must be for all animations!)
	self.curranim = ""
	self.frame = 0
	self.animspeed = 1/16 -- 20
	self.animtimer = self.animspeed
	-- animation step2: create anims
	SPlayables.createAnim(self, g_ANIM_DEFAULT, 1, 7)
	SPlayables.createAnim(self, g_ANIM_IDLE_L, 8, 14) -- L
	SPlayables.createAnim(self, g_ANIM_IDLE_R, 1, 7) -- R
	SPlayables.createAnim(self, g_ANIM_WALK_L, 15, 25)
	SPlayables.createAnim(self, g_ANIM_WALK_R, 26, 36)
	SPlayables.createAnim(self, g_ANIM_RUN_L, 15, 25)
	SPlayables.createAnim(self, g_ANIM_RUN_R, 26, 36)
	SPlayables.createAnim(self, g_ANIM_JUMPUP_L, 15, 25)
	SPlayables.createAnim(self, g_ANIM_JUMPUP_R, 26, 36)
	SPlayables.createAnim(self, g_ANIM_JUMPDOWN_L, 15, 25)
	SPlayables.createAnim(self, g_ANIM_JUMPDOWN_R, 26, 36)
	SPlayables.createAnim(self, g_ANIM_LADDERIDLE_L, 12, 14)
	SPlayables.createAnim(self, g_ANIM_LADDERIDLE_R, 2, 4)
	SPlayables.createAnim(self, g_ANIM_LADDERUP_L, 15, 25)
	SPlayables.createAnim(self, g_ANIM_LADDERUP_R, 26, 36)
	SPlayables.createAnim(self, g_ANIM_LADDERDOWN_L, 15, 25)
	SPlayables.createAnim(self, g_ANIM_LADDERDOWN_R, 26, 36)
	-- the bitmap
	local imgoffsetx, imgoffsety = 0, 0
	local bodyoffsetx, bodyoffsety = -8, -16
	self.bmp = Bitmap.new(self.animsimgs[1]) -- starting bmp texture (from playables class)
	self.bmp:setAnchorPoint(0.5, 0.5)
	self.bmp:setScale(self.scalex, self.scaley)
	self.w = self.bmp:getWidth() / 2 + bodyoffsetx
	self.h = self.bmp:getHeight() + bodyoffsety
	-- set position inside sprite
	self.bmp:setPosition(self.w / 2 + imgoffsetx / 2, self.h / 2 + bodyoffsety / 2)
	-- our sprite is ready
	self.sprite = Sprite.new()
	self.sprite:addChild(self.bmp)
	if xparams2.sdebug then
		local mypixel = Pixel.new(xparams2.sdebug, 0.5, self.w, self.h) -- collisions debugging
		self.sprite:addChild(mypixel) -- debug
		print("* player body:", self.w, self.h)
	end
	-- end
	self:addChild(self.sprite)
	-- listeners
	self:addEventListener(Event.KEY_DOWN, self.onKeyDown, self)
	self:addEventListener(Event.KEY_UP, self.onKeyUp, self)
	self:addEventListener(Event.ENTER_FRAME, self.onEnterFrame, self)
end

-- game loop
function SPlayer1:onEnterFrame(e)
	if self.isdirty and self.currlives > 0 then
		self.cworld.gcam:shake(2, 32) -- GCam:shake(duration, distance)
		local function map(v, minSrc, maxSrc, minDst, maxDst, clampValue)
			local newV = (v - minSrc) / (maxSrc - minSrc) * (maxDst - minDst) + minDst
			return not clampValue and newV or clamp(newV, minDst >< maxDst, minDst <> maxDst)
		end
		-- hud health
		self.currhealth -= 1
		local hudhealthwidth = map(self.currhealth, 0, self.totalhealth, 0, 100)
		self.cworld.hudhealth:setWidth(hudhealthwidth)
		if self.currhealth == 1 then self.cworld.hudhealth:setColor(0xff0000)
		elseif self.currlives == 1 and self.currhealth == 2 then self.cworld.hudhealth:setColor(0xff5500)
		end
		self.washurt = 10 -- hurt timer
		self.sprite:setColorTransform(2, 0.5, 0.5, 1) -- hurt fx
		self.isdirty = false
		if self.currhealth <= 0 then
			-- hud lives
			self.currlives -= 1
			self.cworld.hudlives[self.currlives + 1]:setVisible(false)
			-- hud health
			if self.currlives > 0 then
				self.currhealth = self.totalhealth
				hudhealthwidth = map(self.currhealth, 0, self.totalhealth, 0, 100)
				self.cworld.hudhealth:setWidth(hudhealthwidth)
				self.cworld.hudhealth:setColor(0x00ff00)
				if self.currlives == 1 then self.cworld.hudlives[1]:setColor(0xff0000) end
			else
				self.isdead = true
			end
		end
	end
	if self.isdead then -- deaded
		self.curranim = g_ANIM_LOSE1_R
		self.cworld.losescreen:setPosition(self.x + 30, self.y + 24)
		self.cworld.losescreen:setVisible(true)
		self.sprite:setColorTransform(255*2/255, 255/255, 255/255, 1)
		self.bmp:setY(self.bmp:getY()-2)
		if self.bmp:getY() < -200 then
			self.isdead = false
			scenemanager:changeScene("levelX", 1, transitions[math.random(#transitions)], easings[math.random(#easings)])
		end
	end
	-- actions
	if self.isaction1 then
--		dosound(self.sndshoot)
--		self:shoot()
		Core.asyncCall(self.shoot, self)
	end
end

-- keys handler
function SPlayer1:onKeyDown(e) -- keys pressed
	if self.currlives > 0 then
		-- play sound function
		local function dosound(xsound)
			if self.channel and not self.channel:isPlaying() then
				self.channel = xsound:play()
				if self.channel then self.channel:setVolume(0.1) end
			else
				if self.channel2 and not self.channel2:isPlaying() then
					self.channel2 = xsound:play()
					if self.channel2 then self.channel2:setVolume(0.1) end
				else print("SPlayer1, all channels are busy!")
				end
			end
		end
		-- move
		if e.keyCode == KeyCode.LEFT then self.isleft = true end
		if e.keyCode == KeyCode.RIGHT then self.isright = true end
		if e.keyCode == KeyCode.UP then
--			self.isup = true -- for moving platforms but breaks the ground jump input buffering!
			self.currinputbuffer = self.inputbuffer -- it seems win32 keeps keydown in cache?
		end
		if e.keyCode == KeyCode.DOWN then self.isdown = true end
		-- actions
		if e.keyCode == KeyCode.W or e.keyCode == KeyCode.SPACE then
--			dosound(self.sndshoot)
--			self:shoot()
			Core.asyncCall(self.shoot, self)
		end
	end
end
function SPlayer1:onKeyUp(e) -- keys released
	-- move
	if e.keyCode == KeyCode.LEFT then self.isleft = false end
	if e.keyCode == KeyCode.RIGHT then self.isright = false end
	if e.keyCode == KeyCode.UP then
		if self.vy >= -18 and self.vy < 18 then -- variable jump height???
			self.vy += 4
		else
			self.vy += 8
		end
		self.isup = false -- for moving platforms
		self.wasup = false -- prevent constant jumps
		self.currinputbuffer = 0 -- it seems win32 doesn't release key?
	end
	if e.keyCode == KeyCode.DOWN then
		self.isdown = false
		self.wasdown = false -- prevent constant going down ptpf
	end
	-- restart on key 'R' up
	if e.keyCode == KeyCode.R then -- restart level
--		Core.asyncCall(scenemanager.changeScene, -- multiplies speed bug!
--			scenemanager, "levelX", 1, transitions[math.random(#transitions)], easings[math.random(#easings)])
		scenemanager:changeScene("levelX", 1, transitions[math.random(#transitions)], easings[math.random(#easings)])
	end
end

-- shoot
function SPlayer1:shoot()
	self.isaction1 = false -- cancel shoot touch button
--	local texpath = "gfx/projectiles/Swap-Tap-Anim1.png"
	local texpath = "gfx/projectiles/lazer.png"
	local scalex, scaley = 4, 3
	local speed, jumpspeed = 8*4, 0
	local range = 64*14
	local myshape = SProjectile.new(self.cworld,
		{
			texpath=texpath,
			x=self.x, y=self.y+64,
			r=255/1.5/255, g=255/1.5/255, b=255/4/255, a=1.5,
			scalex=scalex, scaley=scaley,
			flip=self.flip,
			speed=speed, jumpspeed=jumpspeed,
			range=range,
			player1=true, -- missile position settings for player1
--			sdebug=0xffff00,
		}
	)
	myshape.ispprojectile = true -- ECS id, IMPORTANT! XXX
	self.cworld.projectileslayer:addChild(myshape) -- layer
end
