SProjectile = Core.class(Sprite) -- don't need to put in lists

function SProjectile:init(xcworld, xparams)
	self.cworld = xcworld
	-- params
	local params = xparams or {}
	params.scalex = xparams.scalex or 1
	params.scaley = xparams.scaley or params.scalex
	params.flip = xparams.flip or 1
	params.angle = xparams.angle or 0
	params.defaultmass = xparams.defaultmass or 0
	params.speed = xparams.speed or 8*1
	params.jumpspeed = xparams.jumpspeed or 46
	params.animspeed = xparams.animspeed or 1/16 -- 20
	params.dx = xparams.dx or nil
	params.dy = xparams.dy or nil
	params.range = xparams.range or 64*10
	params.dokeep = xparams.dokeep or nil
	params.r = xparams.r or 1
	params.g = xparams.g or 1
	params.b = xparams.b or 1
	params.a = xparams.a or 1
	-- self
	self.x = xparams.x
	self.y = xparams.y
	self.scalex = params.scalex
	self.scaley = params.scaley
	self.flip = params.flip
	self.defaultmass = params.defaultmass
	self.currmass = self.defaultmass
	self.speed = params.speed
	self.jumpspeed = params.jumpspeed
	self.angle = params.angle
	self.vx = 0
	self.vy = 0
	-- ai
	self.startpositionx = params.x
	self.startpositiony = params.y
	self.dx = params.dx
	self.dy = params.dy
	self.range = params.range
	-- the bitmap
	self.bmp = Bitmap.new(Texture.new(params.texpath)) -- starting bmp texture (from playables class)
	self.bmp:setAnchorPoint(0.5, 0.5)
	self.bmp:setScale(self.scalex, self.scaley)
	self.bmp:setRotation(params.angle)
	local bodyoffsetx, bodyoffsety = 0, 0
	self.w = self.bmp:getWidth() / 2 + bodyoffsetx
	self.h = self.bmp:getHeight() / 2 + bodyoffsety
	-- set position inside sprite
	if params.player1 then
		self.bmp:setPosition(0, self.h/2) -- good settings for player1
	else
		self.bmp:setPosition(self.w/2, self.h/2) -- good settings for nmes
	end
	-- our sprite is ready
	self.sprite = Sprite.new()
	self.sprite:addChild(self.bmp)
	if params.sdebug then -- collisions debugging
		local mypixel = Pixel.new(params.sdebug, 0.7, self.w, self.h)
		self.sprite:addChild(mypixel)
	end
	self:addChild(self.sprite)
	self.sprite:setColorTransform(params.r, params.g, params.b, params.a)
	self:setPosition(self.x, self.y)
	-- listeners
	self:addEventListener(Event.ENTER_FRAME, self.onEnterFrame, self)
	-- cbump
	self.cworld:add(self, self.x, self.y, self.w, self.h)
end

-- game loop
local col
local dx, dy
function SProjectile:onEnterFrame(e)
	if self.isdirty and not self.isdeaded then -- destroy
		-- isdirty
		self.isdirty = false
		self.isdeaded = true
		if self.dokeep then self.sprite:setAlpha(0.3)
		else self.sprite:getParent():removeChild(self.sprite) -- remove sprite from layer
		end
		self.cworld:remove(self) -- remove from cbump world
		self = nil -- perfs?
		return
	else
		-- destroy projectiles after a certain distance
		dx, dy = self.startpositionx - self.x, self.startpositiony - self.y
		if (((dx * dx) + (dy * dy)) ^0.5) > self.range then self.isdirty = true end
	end
	if not self.isdeaded then
		-- collisionfilter
		local function collisionfilter(item, other) return "cross" end -- "touch", "cross", "slide", "bounce"
		-- cbump
--		local goalx = self.x + self.vx
--		local goaly = self.y + self.vy
		local goalx = self.x + self.speed * self.flip
--		local goaly = self.y + self.jumpspeed
		local goaly = self.y + self.jumpspeed + self.vy
		local nextx, nexty, collisions, len = self.cworld:move(self, goalx, goaly, collisionfilter)
		-- collisions
		for i = 1, len do
			col = collisions[i]
			if col.item.ispprojectile and not -- player projectile
				(col.other.isptpf or col.other.isplayer1 or col.other.iscollectible) then
					col.item.isdirty = true
					col.other.isdirty = true
--					if col.other.isnme then dosound(self.sndnmeAhurt) end
			end
			if col.item.iseprojectile and not -- nme projectile
				(col.other.isptpf or col.other.isnme or col.other.iseprojectile or col.other.iscollectible) then
					col.item.isdirty = true
					col.other.isdirty = true
--					if col.other.isplayer1 then dosound(self.sndplayer1hurt) end
			end
		end
		-- physics
		self.vy += g_gravity * self.currmass
		-- move sprite & scale & flip
		self.x, self.y = nextx, nexty
		self:setPosition(self.x, self.y)
	end
end
