Tiled_Levels = Core.class(Sprite)

function Tiled_Levels:init(xcworld, xtiledlevel)
	-- Tiled map size (parallax, ...)
	self.mapwidth = xtiledlevel.width * xtiledlevel.tilewidth
	self.mapheight = xtiledlevel.height * xtiledlevel.tileheight
	-- level sprite will hold the whole level graphics in layers and will serve as a "camera"
	self.camera = Sprite.new()
	-- let's go!
--	local sdebug = 0xff00ff -- uncomment for collision debugging
	-- _   _ _                _   _                                 
	--| | (_) |              | | (_)                                
	--| |_ _| | ___  ___  ___| |_ _ _ __ ___   __ _  __ _  ___  ___ 
	--| __| | |/ _ \/ __|/ _ \ __| | '_ ` _ \ / _` |/ _` |/ _ \/ __|
	--| |_| | |  __/\__ \  __/ |_| | | | | | | (_| | (_| |  __/\__ \
	-- \__|_|_|\___||___/\___|\__|_|_| |_| |_|\__,_|\__, |\___||___/
	--                                               __/ |          
	--                                              |___/           
	local tilesetimages = {} -- the table holding all Tiled images info (path, width, height)
	local tilesets = xtiledlevel.tilesets
	for i = 1, #tilesets do
		local tileset = tilesets[i]
		local tiles = tileset.tiles
		local path
		for j = 1, #tiles do
			if application:getDeviceInfo() == "Web" then
				if tiles[j].image:match("../../") then -- file is outside "tiled/levels/" folder
					path=tiles[j].image:gsub("../../", "")
				else -- file is inside "tiled/levels/" folder
					path="tiled/levels/"..tiles[j].image
				end
			else -- desktop (will automatically switch to correct folder (../../)
				path="tiled/levels/"..tiles[j].image
			end
			tilesetimages[tileset.firstgid + tiles[j].id] = {
				path=path,
				width=tiles[j].width,
				height=tiles[j].height,
			}
		end
	end
	--                           _____                            
	--                          |_   _|                           
	-- _ __   __ _ _ __ ___  ___  | |  _ __ ___   __ _  __ _  ___ 
	--| '_ \ / _` | '__/ __|/ _ \ | | | '_ ` _ \ / _` |/ _` |/ _ \
	--| |_) | (_| | |  \__ \  __/_| |_| | | | | | (_| | (_| |  __/
	--| .__/ \__,_|_|  |___/\___|_____|_| |_| |_|\__,_|\__, |\___|
	--| |                                               __/ |     
	--|_|                                              |___/      
	-- parse Tiled layers images
	local function parseImage(xobject, xlayer)
		local tex = Texture.new(tilesetimages[xobject.gid].path, false)
		local bitmap = Bitmap.new(tex)
		bitmap:setAnchorPoint(0, 1) -- because I always forget to modify Tiled objects alignment
		-- support for Tiled image scaling
		local scalex, scaley = xobject.width / tex:getWidth(), xobject.height / tex:getHeight()
		bitmap:setScale(scalex, scaley)
		bitmap:setPosition(xobject.x, xobject.y)
		bitmap:setRotation(xobject.rotation) -- image rotated in Tiled?
		-- addChild
		xlayer:addChild(bitmap)
		bitmap = nil
	end
	-- __ _     __
	--/__|_||V||_ 
	--\_|| || ||__
	-- makeDeco
	local function makeDeco(xobject, xtable, xlevelsetup, xspritelayer)
		local levelsetup = xlevelsetup or {}
		if xtable then
			for k, v in pairs(xtable) do levelsetup[k] = v end
		end
		local myshape = self:buildShapes(xobject, levelsetup)
		myshape:setPosition(xobject.x, xobject.y)
		xspritelayer:addChild(myshape)
	end
	--  _____        _____   _____ ______   _______ _    _ ______   _      ________      ________ _      
	-- |  __ \ /\   |  __ \ / ____|  ____| |__   __| |  | |  ____| | |    |  ____\ \    / /  ____| |     
	-- | |__) /  \  | |__) | (___ | |__       | |  | |__| | |__    | |    | |__   \ \  / /| |__  | |     
	-- |  ___/ /\ \ |  _  / \___ \|  __|      | |  |  __  |  __|   | |    |  __|   \ \/ / |  __| | |     
	-- | |  / ____ \| | \ \ ____) | |____     | |  | |  | | |____  | |____| |____   \  /  | |____| |____ 
	-- |_| /_/    \_\_|  \_\_____/|______|    |_|  |_|  |_|______| |______|______|   \/   |______|______|
	-- Tiled layers
	local layers = xtiledlevel.layers
	for i = 1, #layers do
		local group = layers[i] -- current group (folder)
		local grouplayers -- layers in group
		local layer -- individual layer (in layers or grouplayers)
		local objects -- objects in layer
		local object -- individual object
		local myshape -- object shape (added once)
		local mytable -- per shape table
		local levelsetup -- common shapes settings (table)
		-- _____ _   ___      _______  _____ _____ ____  _      ______   __  __  _____ 
		--|_   _| \ | \ \    / /_   _|/ ____|_   _|  _ \| |    |  ____| |  \/  |/ ____|
		--  | | |  \| |\ \  / /  | | | (___   | | | |_) | |    | |__    | \  / | |  __ 
		--  | | | . ` | \ \/ /   | |  \___ \  | | |  _ <| |    |  __|   | |\/| | | |_ |
		-- _| |_| |\  |  \  /   _| |_ ____) |_| |_| |_) | |____| |____  | |  | | |__| |
		--|_____|_| \_|   \/   |_____|_____/|_____|____/|______|______| |_|  |_|\_____|
		if group.name == "INVISIBLE_MG" then -- Tiled group (folder)
			-- transparent object, only for collisions!
			grouplayers = group.layers
			for j = 1, #grouplayers do
				layer = grouplayers[j]
				if layer.name == "inv_physics_steps" then
					objects = layer.objects
					for j = 1, #objects do
						object = objects[j]
						levelsetup = {} -- reset
						if object.name == "" then
							mytable = {}
						end
						if mytable then
							levelsetup = {} -- params common to all objects
							for k, v in pairs(mytable) do levelsetup[k] = v end
							myshape = self:buildShapes(object, levelsetup)
							myshape:setPosition(object.x, object.y)
							myshape.isstep = true -- id, IMPORTANT! XXX
							xcworld:add(myshape, object.x, object.y, myshape.w, myshape.h)
						end
					end
				-- __ _  _  _  _ 
				--|_ |_)|_)/ \|_)
				--|__| \| \\_/| \
				else print("* error layer: ", layer.name) end
			end
		-- ______ _____ 
		--|  ____/ ____|
		--| |__ | |  __ 
		--|  __|| | |_ |
		--| |   | |__| |
		--|_|    \_____|
		elseif group.name == "FG" then -- Tiled group (folder)
			grouplayers = group.layers
			for i = 1, #grouplayers do
				layer = grouplayers[i]
				--  __             _                 _                                 
				-- / _|           | |               (_)                                
				--| |_ __ _     __| | ___  ___ ___   _ _ __ ___   __ _  __ _  ___  ___ 
				--|  _/ _` |   / _` |/ _ \/ __/ _ \ | | '_ ` _ \ / _` |/ _` |/ _ \/ __|
				--| || (_| |  | (_| |  __/ (_| (_) || | | | | | | (_| | (_| |  __/\__ \
				--|_| \__, |   \__,_|\___|\___\___/ |_|_| |_| |_|\__,_|\__, |\___||___/
				--     __/ |_____               ______                  __/ |          
				--    |___/______|             |______|                |___/           
				if layer.name == "fg_deco_images" then
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for j = 1, #objects do parseImage(objects[j], self[layer.name]) end
				--  __             _                     _                           
				-- / _|           | |                   | |                          
				--| |_ __ _     __| | ___  ___ ___   ___| |__   __ _ _ __   ___  ___ 
				--|  _/ _` |   / _` |/ _ \/ __/ _ \ / __| '_ \ / _` | '_ \ / _ \/ __|
				--| || (_| |  | (_| |  __/ (_| (_) |\__ \ | | | (_| | |_) |  __/\__ \
				--|_| \__, |   \__,_|\___|\___\___/ |___/_| |_|\__,_| .__/ \___||___/
				--     __/ |_____               ______              | |              
				--    |___/______|             |______|             |_|              
				elseif layer.name == "fg_deco_shapes" then
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for i = 1, #objects do
						object = objects[i]
						if object.name == "" then
							mytable = {
								texpath="tiled/levels/grounds/Abstract_004_SPEC.jpg",
								r=126/255, g=91/255, b=83/255, a=1,
								scalex=0.7,
							}
						end
						levelsetup = {} -- params common to all objects
						makeDeco(object, mytable, levelsetup, self[layer.name])
					end
				end
			end
		-- __  __  _____ 
		--|  \/  |/ ____|
		--| \  / | |  __ 
		--| |\/| | | |_ |
		--| |  | | |__| |
		--|_|  |_|\_____|
		elseif group.name == "MG" then -- Tiled group (folder)
			grouplayers = group.layers
			for i = 1, #grouplayers do
				layer = grouplayers[i]
				-- __    _  _  _ 
				--|_ |  / \/ \|_)
				--|  |__\_/\_/| \
				if layer.name == "mg_floor" then
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for i = 1, #objects do
						object = objects[i]
						myshape, mytable = nil, nil -- reset
						levelsetup = {} -- reset
						if object.name == "" then -- default
							mytable = {
								texpath="tiled/levels/grounds/Abstract_004_SPEC.png",
								scalex=3,
								r=200/255, g=62/255, b=108/255, alpha=1,
							}
						elseif object.name == "ground" then
							mytable = {
								texpath="tiled/levels/grounds/Abstract_004_SPEC.png",
								scalex=3,
								r=127/255, g=62/255, b=108/255, alpha=1,
							}
						end
						if mytable then
							levelsetup = {}
							for k, v in pairs(mytable) do levelsetup[k] = v end
							myshape = self:buildShapes(object, levelsetup)
							myshape:setPosition(object.x, object.y)
							myshape.isfloor = true
							self[layer.name]:addChild(myshape)
							xcworld:add(myshape, object.x, object.y, myshape.w, myshape.h)
						end
						-- cleaning?
						myshape = nil
					end
				-- _  _____ _  _  __
				--|_|/   | / \|_)(_ 
				--| |\__ | \_/| \__)
				elseif layer.name == "mg_actors" then
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for i = 1, #objects do
						object = objects[i]
						myshape, mytable = nil, nil
						-- _     _ \ / __ _       
						--|_)|  |_| Y |_ |_)   /| 
						--|  |__| | | |__| \    | 
						if object.name == "player1" then
							xcworld.player1 = SPlayer1.new(xcworld,
								{
									spritesheetpath="gfx/players/Sci-fi_Soldiers_2_military_0001c.png", cols=6, rows=6,
									x=object.x, y=object.y,
								},
								{
									-- all paramaters are set directly in SPlayer1 class
									sdebug=sdebug,
								}
							)
							xcworld.player1:setPosition(object.x, object.y)
							self[layer.name]:addChild(xcworld.player1)
							xcworld:add(xcworld.player1, object.x, object.y, xcworld.player1.w, xcworld.player1.h)
						-- __ _        __ ________ _     __ __
						--/  / \|  |  |_ /   |  | |_)|  |_ (_ 
						--\__\_/|__|__|__\__ | _|_|_)|__|____)
						elseif object.name:find("^coin") then
							myshape = SCollectible.new(xcworld,
								{
									spritesheetpath="gfx/collectibles/coin_20_x01.png", cols=6, rows=1,
									version = object.name,
									x=object.x, y=object.y,
									scalex=2,
									hasanims=true,
									animations={
										{g_ANIM_DEFAULT, 1, 5},
									},
									animspeed=1/12,
									sdebug=sdebug,
								}
							)
							self[layer.name]:addChild(myshape)
							xcworld:add(myshape, object.x, object.y, myshape.w, myshape.h)
						end
					end
				end
			end
		-- ____   _____ 
		--|  _ \ / ____|
		--| |_) | |  __ 
		--|  _ <| | |_ |
		--| |_) | |__| |
		--|____/ \_____|
		elseif group.name == "BG" then -- Tiled group (folder)
			grouplayers = group.layers
			for i = 1, #grouplayers do
				layer = grouplayers[i]
				-- _                         _                 
				--| |                       | |                
				--| |__   __ _     __ _  ___| |_ ___  _ __ ___ 
				--| '_ \ / _` |   / _` |/ __| __/ _ \| '__/ __|
				--| |_) | (_| |  | (_| | (__| || (_) | |  \__ \
				--|_.__/ \__, |   \__,_|\___|\__\___/|_|  |___/
				--        __/ |_____                           
				--       |___/______|                          
				if layer.name == "bg_actors" then
					self[layer.name] = self[layer.name] or Sprite.new() -- 20230526 XXX
					self.camera:addChild(self[layer.name]) -- 20230526 XXX
					xcworld.mg_projectiles = self[layer.name] -- for the bullets layer
					objects = layer.objects
					for j = 1, #objects do
						object = objects[j]
						myshape, mytable = nil, nil
						--               _ _    _                                           
						--              | | |  (_)                                          
						--__      ____ _| | | ___ _ __   __ _     _ __  _ __ ___   ___  ___ 
						--\ \ /\ / / _` | | |/ / | '_ \ / _` |   | '_ \| '_ ` _ \ / _ \/ __|
						-- \ V  V / (_| | |   <| | | | | (_| |   | | | | | | | | |  __/\__ \
						--  \_/\_/ \__,_|_|_|\_\_|_| |_|\__, |   |_| |_|_| |_| |_|\___||___/
						--                               __/ |_____                         
						--                              |___/______|                        
						if object.name:match("^walking_nme") then -- notice the ^ (begining match)
							-- walking_nmeAl_R416_S80
							myshape = nil -- reset
							local range
							local speed
							local index
							index = object.name:find("_R")
							if index then
								range = tonumber(object.name:sub(index):match("%d+")) -- find Rnnn
							end
							index = object.name:find("_S")
							if index then speed = tonumber(object.name:sub(index):match("%d+")) / 100 -- find Snnn
							else speed = 1
							end
							local version = object.name:sub(12, 13) -- Ax, Bx, ..., x is the direction (a, u, d, l, r)
							if version:sub(1, 1) == "C" then
								myshape = SWalkingNme.new(xcworld,
									{
										spritesheetpath="gfx/nmes/100Avatars_007_Observer_181_anims_0001.png", cols=4, rows=5,
										x=object.x, y=object.y,
									},
									{
										breed=version:sub(1, 1),
										scalex=1.6, scaley=-1.6,
										defaultmass=-1,
										totallives=3,
										speed=speed, jumpspeed=8*16,
										coyotetimer=12,
										animations={
											{g_ANIM_DEFAULT, 1, 20},
											{g_ANIM_RUN_R, 1, 20},
										},
										animspeed=1/speed/8,
										dx=range,
										dokeep=true,
										direction=version:sub(2, 2),
										sdebug=sdebug,
									}
								)
							end
							if myshape then
								myshape:setPosition(object.x, object.y)
								self[layer.name]:addChild(myshape)
								xcworld.walkingnmes[#xcworld.walkingnmes + 1] = myshape
								xcworld:add(myshape, object.x, object.y, myshape.w, myshape.h)
							end
						end
					end
				-- _                  _                 _                                      _ 
				--| |                | |               (_)                                    | |
				--| |__   __ _     __| | ___  ___ ___   _ _ __ ___   __ _  __ _  ___  ___     | |__ 
				--| '_ \ / _` |   / _` |/ _ \/ __/ _ \ | | '_ ` _ \ / _` |/ _` |/ _ \/ __|    | '_ \
				--| |_) | (_| |  | (_| |  __/ (_| (_) || | | | | | | (_| | (_| |  __/\__ \    | |_) |
				--|_.__/ \__, |   \__,_|\___|\___\___/ |_|_| |_| |_|\__,_|\__, |\___||___/    |_.__/...
				--        __/ |_____               ______                  __/ |       ______ 
				--       |___/______|             |______|                |___/       |______|
				elseif layer.name:match("^bg_deco_images_") then -- notice the ^ (begining match)
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for j = 1, #objects do parseImage(objects[j], self[layer.name]) end
				-- _                  _                     _                           
				--| |                | |                   | |                          
				--| |__   __ _     __| | ___  ___ ___   ___| |__   __ _ _ __   ___  ___ 
				--| '_ \ / _` |   / _` |/ _ \/ __/ _ \ / __| '_ \ / _` | '_ \ / _ \/ __|
				--| |_) | (_| |  | (_| |  __/ (_| (_) |\__ \ | | | (_| | |_) |  __/\__ \
				--|_.__/ \__, |   \__,_|\___|\___\___/ |___/_| |_|\__,_| .__/ \___||___/
				--        __/ |_____               ______              | |              
				--       |___/______|             |______|             |_|              
				elseif layer.name == "bg_deco_shapes" then
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for i = 1, #objects do
						object = objects[i]
						if object.name == "" then
							mytable = {
								texpath="tiled/levels/grounds/Abstract_004_SPEC.jpg",
								r=126/2/255, g=91/2/255, b=83/2/255, a=1,
								scalex=0.6,
							}
						elseif object.name == "direction" then
							mytable = {
								color=0x55ffff,
								shapelinewidth=5, shapelinecolor=0x287878,
							}
						end
						levelsetup = {} -- params common to all objects
						makeDeco(object, mytable, levelsetup, self[layer.name])
					end
				-- _                       _         _       _    __                         
				--| |                     | |       | |     | |  / _|                        
				--| |__   __ _       _ __ | |_ _ __ | | __ _| |_| |_ ___  _ __ _ __ ___  ___ 
				--| '_ \ / _` |     | '_ \| __| '_ \| |/ _` | __|  _/ _ \| '__| '_ ` _ \/ __|
				--| |_) | (_| |     | |_) | |_| |_) | | (_| | |_| || (_) | |  | | | | | \__ \
				--|_.__/ \__, |     | .__/ \__| .__/|_|\__,_|\__|_| \___/|_|  |_| |_| |_|___/
				--        __/ |_____| |       | |                                            
				--       |___/______|_|       |_|                                            
				elseif layer.name == "bg_ptplatforms" then
					self[layer.name] = self[layer.name] or Sprite.new()
					self.camera:addChild(self[layer.name])
					objects = layer.objects
					for i = 1, #objects do
						object = objects[i]
						myshape, mytable = nil, nil -- reset
						levelsetup = {} -- reset
						if object.name == "" then -- transparent
							mytable = {
								texpath="tiled/levels/grounds/Abstract_004_SPEC.png",
								r=126/2/255, g=91/2/255, b=83/2/255, a=1,
								scalex=3,
							}
						end
						levelsetup = {}
						for k, v in pairs(mytable) do levelsetup[k] = v end
						myshape = self:buildShapes(object, levelsetup)
						myshape:setPosition(object.x, object.y)
						myshape.isptpf = true -- id, IMPORTANT! XXX
						self[layer.name]:addChild(myshape)
						xcworld:add(myshape, object.x, object.y, myshape.w, myshape.h)
						-- cleaning?
						myshape = nil
					end
				end
			end
		end
	end
end

-- BUILDSHAPES
function Tiled_Levels:buildShapes(xobject, xlevelsetup)
	local myshape = nil
	local tablebase = {}
	if xobject.shape == "ellipse" then
		tablebase = {
			x=xobject.x, y=xobject.y,
			w=xobject.width, h=xobject.height, rotation=xobject.rotation,
		}
		for k, v in pairs(xlevelsetup) do tablebase[k] = v end
		myshape = Tiled_Shape_Ellipse.new(tablebase)
	elseif xobject.shape == "polygon" then
		tablebase = {
			x=xobject.x, y=xobject.y,
			coords=xobject.polygon, rotation=xobject.rotation,
		}
		for k, v in pairs(xlevelsetup) do tablebase[k] = v end
		myshape = Tiled_Shape_Polygon.new(tablebase)
	elseif xobject.shape == "polyline" then
		tablebase = {
			x=xobject.x, y=xobject.y,
			coords=xobject.polyline, rotation=xobject.rotation,
		}
		for k, v in pairs(xlevelsetup) do tablebase[k] = v end
		myshape = Tiled_Shape_Polygon.new(tablebase)
	elseif xobject.shape == "rectangle" then
		tablebase = {
			x=xobject.x, y=xobject.y,
			w=xobject.width, h=xobject.height, rotation=xobject.rotation,
		}
		for k, v in pairs(xlevelsetup) do tablebase[k] = v end
		myshape = Tiled_Shape_Rectangle.new(tablebase)
	else
		print("*** CANNOT PROCESS THIS SHAPE! ***", xobject.shape, xobject.name)
		return
	end

	return myshape
end
